#GiD-Plugin to create voronoi polygons using voro++

namespace eval Voro {         
    variable info_plugin ;#information read from .xml plugin file
    set info_plugin(path) [file dirname [info script]] ;#to know where to find the files
    if { [info exists info_plugin(Version)] } {
        set other_plugin_version $info_plugin(Version)
    }
    array set info_plugin [ReadProblemtypeXml [file join $info_plugin(path) voro.xml] InfoPlugin {Name Version MinimumGiDVersion}]  
    if { [info exists other_plugin_version] && $other_plugin_version != $info_plugin(Version)} {
        WarnWinText [_ "There are two versions of %s: the %s and %s inside the plugins folder. One must be deleted" $info_plugin(Name) $other_plugin_version $info_plugin(Version)]
    }
    if { [GidUtils::VersionCmp $info_plugin(MinimumGiDVersion)] < 0 } {  
        WarnWinText [_ "This plugin requires GiD %s or later" $info_plugin(MinimumGiDVersion)]
    } 
    variable options
}

proc Voro::OptionsReset { } {   
    variable options
    set options(dimension) 3 ;#2 2.5 or 3
    set options(random) 1
    set options(num_random_centers) 10
    set options(save_random_centers) 0
    set options(particles_filename) ""
    set options(x_min) 0.0
    set options(x_max) 1.0
    set options(y_min) 0.0
    set options(y_max) 1.0
    set options(z_min) 0.0
    set options(z_max) 1.0 
    set options(x_periodic) 0
    set options(y_periodic) 0
    set options(z_periodic) 0
    set options(radical) 0
    set options(r_min) 0.5
    set options(r_max) 1.5 
    set options(share) 1
    set options(collapse) 0
    set options(tolerance) 1e-3
    set options(create_contacts_periodicity) 1
    set options(create_contacts_periodicity_element_type) None ;#None Line or Prism
    set options(create_contacts_inner) 0
    set options(periodic_box) 1 ;#else the periodic shape won't be a box!!
    #create matrix current implementation is probably incompatible with other options !!
    set options(create_matrix) 0 ;#create matrix scaling grains
    set options(create_matrix_delete_outer) 0 ;#to delete the matrix touching the boundary
    set options(create_matrix_create_grain) 0 ;#create also the grain volume inside (volume original=volume grain+volumes matrix)
    set options(create_matrix_scale_factor) 0.93 ;#approx volume 15%
    #                                             approx volume: 0.85-> 24% , 0.91->15%, 0.94 -> 10%
    #if !$options(create_matrix_delete_outer) then approx volume: 0.93-> 20% , 0.95->14%
    set options(classify_groups) 1 ;#apply groups to parts
    
    #internal only
    set options(convert_surfaces_to_nurbs) 1 ;#they are meshed better as NURBS because planar are obsolete
    set options(try_fix_gid_bug_first_verso_1) 1 ;#to try to avoid create a planar surface with first curve orientation==1  
    set options(local_collapse_by_volume) 0
}

proc Voro::GetExe { } {
    variable info_plugin
    if { $::tcl_platform(pointerSize) == 8} {
        set bits 64
    } else {
        set bits 32
    }
    if { $::tcl_platform(platform) == "windows"} {
        set exe [file join $info_plugin(path) bin win-x${bits} voro++.exe]
    } else {
        set exe [file join $info_plugin(path) bin linux-x${bits} voro++]
    }
    return $exe
}

proc Voro::GenerateRandomParticles { filename } {
    variable options
    set fp [open $filename w]
    if { $fp != "" } {
        foreach axis {x y z} {
            if { $options(${axis}_periodic) } {
                #decrease the region of point a 10% to allow periodic shape
                set d [expr {$options(${axis}_max)-$options(${axis}_min)}]
                set limit(${axis}_min) [expr {$options(${axis}_min)+0.05*$d}]
                set limit(${axis}_max) [expr {$options(${axis}_max)-0.05*$d}]
            } else {
                set limit(${axis}_min) $options(${axis}_min)
                set limit(${axis}_max) $options(${axis}_max)        
            } 
        }                
        for {set i 1} {$i <= $options(num_random_centers)} {incr i} {
            foreach axis {x y z} {
                set v [expr {rand()}]
                set p($axis) [expr {$v*$limit(${axis}_max)+(1.0-$v)*$limit(${axis}_min)}]
            }
            if { $options(dimension) != 3 } {
                #dimension 2.5 mean a 3D shape but prismatic (extrussion of a 2D shape)
                set p(z) 0.0
            }
            if { $options(radical) } {
                set v [expr {rand()}]
                set r [expr {$v*$options(r_max)+(1.0-$v)*$options(r_min)}]
                puts $fp "$i $p(x) $p(y) $p(z) $r"
            } else {
                puts $fp "$i $p(x) $p(y) $p(z)"
            }
        }
        close $fp
    }
}

proc Voro::CreateExtrudedEntitiesScalingInside { } {
    variable options    
    if { $options(dimension) == 2 } {
        set entity_type_to_extrude surface
        set entity_type_to_copy line
        set entity_type_lower point
    } else {
        set entity_type_to_extrude volume        
        set entity_type_to_copy surface
        set entity_type_lower line
    }
    #codigo a saco, asumiendo que solo hay los volumenes esperados y empiezan por 1    
    set volume_ids [GiD_Geometry list $entity_type_to_extrude 1:end]
    set first_extruded_volume_id [expr {[GiD_Info geometry maxnum${entity_type_to_extrude}s]+1}]
    
    set master_surface_ids [list]
    foreach volume_id $volume_ids {
        set scale_center [Voro::GetEntityCenter $entity_type_to_extrude $volume_id]
        set volume_surface_ids [Voro::GetEntityLowerEntities $entity_type_to_extrude $volume_id]
        if { $options(create_matrix_create_grain) } {
            #must know the new surfaces that define the volume grain
            set next_volume_id [expr {[GiD_Info geometry maxnum${entity_type_to_extrude}s]+1}]
        }
        GiD_Process Mescape Utilities Copy ${entity_type_to_copy}s Duplicate DoExtrude ${entity_type_to_extrude}s MaintainLayers Scale FNoJoin {*}$scale_center \
            FNoJoin $options(create_matrix_scale_factor) $options(create_matrix_scale_factor) $options(create_matrix_scale_factor) \
            {*}$volume_surface_ids escape
        if { $options(create_matrix_create_grain) } {           
            array unset extruded_volume_slave_surface
            set local_extruded_volume_ids [GiD_Geometry list $entity_type_to_extrude $next_volume_id:end]
            foreach extruded_volume_id $local_extruded_volume_ids {
                set extruded_volume_surface_ids [Voro::GetEntityLowerEntities $entity_type_to_extrude $extruded_volume_id]
                if {$entity_type_to_extrude == "volume" } {
                    #first surface was the master and last the copied slave
                    set extruded_volume_slave_surface([lindex $extruded_volume_surface_ids 0]) [lindex $extruded_volume_surface_ids end]
                } elseif {$entity_type_to_extrude == "surface" } {
                    #first line was the master and thirrd the copied slave
                    set extruded_volume_slave_surface([lindex $extruded_volume_surface_ids 0]) [lindex $extruded_volume_surface_ids 2]
                } else {
                    W "Voro::CreateExtrudedVolumesScalingInsid. unexpected $entity_type_to_extrude"
                }
            }
            #replace the volume by the inner volume (grain)
            set data [GiD_Geometry get $entity_type_to_extrude $volume_id]
            if { $entity_type_to_extrude == "volume" } {
                set new_data [lrange $data 0 1]
                foreach item [lrange $volume_data 2 end] {
                    lset item 0 $extruded_volume_slave_surface([lindex $item 0])
                    lappend new_data $item
                }
            } else {
                set num_lines [lindex $data 2]
                if { [lindex $data 0] == "nurbssurface" } {
                    set first_line_index 9
                } elseif { [lindex $data 0] == "planarsurface" } {
                    set first_line_index 3
                } else {
                    set first_line_index 3
                }
                set new_data [lrange $data 0 $first_line_index-1]
                foreach item [lrange $data $first_line_index [expr {$first_line_index+$num_lines-1}]] {
                    lset item 0 $extruded_volume_slave_surface([lindex $item 0])
                    lappend new_data $item
                }
                lappend new_data {*}[lrange $data [expr {$first_line_index+$num_lines}] end]
            }
            GiD_Geometry delete $entity_type_to_extrude $volume_id
            GiD_Geometry create $entity_type_to_extrude $volume_id {*}$new_data
        } else {
            GiD_Geometry delete $entity_type_to_extrude $volume_id
        }
        lappend master_surface_ids {*}$volume_surface_ids
    }
    set created_volume_ids [GiD_Geometry list $entity_type_to_extrude $first_extruded_volume_id:end]
    set created_surfaces_ids [list]
    foreach volume_id $created_volume_ids {
        lappend created_surfaces_ids {*}[Voro::GetEntityLowerEntities $entity_type_to_extrude $volume_id]
    }
    set created_surfaces_ids [lsort -integer -unique $created_surfaces_ids]
    
    if {$entity_type_to_extrude == "volume" } {
        set num_divisions 2
        set master_surface_ids [lsort -integer -unique $master_surface_ids]    
        GiD_Process Mescape Meshing SemiStructured Volumes $num_divisions {*}$created_volume_ids escape escape
        GiD_Process Mescape Meshing SemiStructured SetMaster {*}$master_surface_ids escape escape
        #GiD_Process Mescape Meshing ElemType Quadrilateral {*}$created_surfaces_ids escape
        #GiD_Process Mescape Meshing ElemType Hexahedra {*}$created_volume_ids escape
    } else {
        #structured surfaces?
    }
    
    if { $options(create_matrix_delete_outer) } {        
        #delete the outer matrix volumes
        foreach surface_id $master_surface_ids {
            set surface_volume_ids [Voro::GetEntityHigherEntities $entity_type_to_copy $surface_id]
            if { [llength $surface_volume_ids] == 1 } {
                set outer_surface($surface_volume_ids) $surface_id
            }
        }        
        set outer_volumes [lsort -integer [array names outer_surface]]
        foreach volume_id $outer_volumes {
            set surface_id $outer_surface($volume_id)
            #must delete volume (and lowerentities) and move the points of the opposite surface to be on surface_id
        }
        GiD_Process Mescape Geometry Delete ${entity_type_to_extrude}s LowerEntities {*}$outer_volumes escape
    }
    
    if { $options(classify_groups) } {
        array set axis_color { x #ff0000ff y #00ff00ff z #0000ffff }
        if { $options(dimension) == 2 } {
            set axes {x y}
        } else {
            set axes {x y z}
        }
        foreach axis $axes {
            GiD_Groups create ${axis}_min_${entity_type_to_copy}s
            GiD_Groups edit color ${axis}_min_${entity_type_to_copy}s $axis_color($axis)
            GiD_Groups create ${axis}_max_${entity_type_to_copy}s
            GiD_Groups edit color ${axis}_max_${entity_type_to_copy}s $axis_color($axis)
        }
        GiD_Groups create inner_${entity_type_to_copy}s
        GiD_Groups edit color inner_${entity_type_to_copy}s #ffff00ff
        GiD_Groups create outer_${entity_type_to_copy}s
        GiD_Groups edit color outer_${entity_type_to_copy}s #00ffffff
        set box_surface_ids [list]
        if { $options(create_matrix_delete_outer) } {            
            set line_ids [GiD_Geometry list $entity_type_lower 1:end]
            foreach line_id $line_ids {
                set line_center [Voro::GetEntityCenter $entity_type_lower $line_id]
                set line_groups [list]
                foreach center_coordinate $line_center axis {x y z} {
                    if { $options(dimension) == 2 && $axis == "z" } {
                        continue
                    }
                    if { [expr {abs($center_coordinate-$options(${axis}_min))}]<1e-8 } {
                        lappend line_groups ${axis}_min_${entity_type_to_copy}s
                    } elseif { [expr {abs($center_coordinate-$options(${axis}_max))}]<1e-8 } {
                        lappend line_groups ${axis}_max_${entity_type_to_copy}s
                    }
                }
                if { [llength $line_groups] } {
                    set line_surface_ids [Voro::GetEntityHigherEntities $entity_type_lower $line_id]
                    set line_surface_boundary_ids [list]
                    foreach surface_id $line_surface_ids {
                        if { [llength [Voro::GetEntityHigherEntities $entity_type_to_copy $surface_id]] == 1 } {
                            lappend line_surface_boundary_ids $surface_id
                        }
                    }
                    foreach group $line_groups {
                        GiD_EntitiesGroups assign $group ${entity_type_to_copy} $line_surface_boundary_ids
                    }   
                    lappend box_surface_ids {*}$line_surface_boundary_ids           
                }            
            }  
            set box_surface_ids [lsort -integer -unique $box_surface_ids]            
        } else {
            #boundary matriz not deleted
            foreach surface_id $master_surface_ids {
                set surface_volume_ids [Voro::GetEntityHigherEntities $entity_type_to_copy $surface_id]
                if { [llength $surface_volume_ids] == 1 } {
                    set surface_center [Voro::GetEntityCenter $entity_type_to_copy $surface_id]
                    set surface_groups [list]
                    foreach center_coordinate $surface_center axis {x y z} {
                        if { $options(dimension) == 2 && $axis == "z" } {
                            continue
                        }
                        if { [expr {abs($center_coordinate-$options(${axis}_min))}]<1e-8 } {
                            lappend surface_groups ${axis}_min_${entity_type_to_copy}s
                        } elseif { [expr {abs($center_coordinate-$options(${axis}_max))}]<1e-8 } {
                            lappend surface_groups ${axis}_max_${entity_type_to_copy}s
                        }
                    }
                    if { [llength $surface_groups] } {                        
                        foreach group $surface_groups {
                            GiD_EntitiesGroups assign $group ${entity_type_to_copy}s $surface_id
                        }
                        lappend box_surface_ids $surface_id
                    }                    
                }
            }            
        }
        
        set inner_surface_ids [list]
        set outer_surface_ids [list]
        set surface_ids [GiD_Geometry list $entity_type_to_copy 1:end]           
        foreach surface_id $surface_ids {
            set surface_higherentity [llength [Voro::GetEntityHigherEntities $entity_type_to_copy $surface_id]]
            if { $surface_higherentity == 1 } {
                lappend outer_surface_ids $surface_id
            } elseif { $surface_higherentity == 2 } {
                lappend inner_surface_ids $surface_id
            } else {
                W "unexpected surface $surface_id belonging to $surface_higherentity volumes"
            }
        }       
        set outer_surface_2_ids [list] ;#outer except box-touching surfaces
        foreach surface_id $outer_surface_ids {
            if { [lsearch -integer -sorted $box_surface_ids $surface_id] == -1 } {
                lappend outer_surface_2_ids $surface_id
            }
        }        
        GiD_EntitiesGroups assign inner_${entity_type_to_copy}s ${entity_type_to_copy}s $inner_surface_ids
        GiD_EntitiesGroups assign outer_${entity_type_to_copy}s ${entity_type_to_copy}s $outer_surface_2_ids
    }
        
}

proc Voro::CreateAsk { } {
    variable options
    set do_it 1
    if { [GiD_Info geometry maxnumpoints] } {
        set ret [tk_dialogRAM  .gid.modal [= "Warning"] [format %s%s%s [= "The model has some entity"]. \n [= "Start a new model?"]] question 0 [_ "Ok"] [_ "Cancel"]]
        if { $ret == 0 } {
            GiD_Process MEscape Files New No escape
        } else {
            set do_it 0
        }
    }
    if { $do_it } {
        GidUtils::DisableGraphics
        if { [catch { Voro::Create } msg] } {
            W $msg
        }
        GidUtils::EnableGraphics
        
        GiD_Process 'Zoom Frame
        GidUtils::UpdateWindow LAYER
        if { $options(classify_groups) } {
            GidUtils::UpdateWindow GROUPS
        }
        if { $options(create_matrix) }  {
            if { $options(dimension) == 2 } {
                set entity_type surface
                set entity_type_lower line
            } else {
                set entity_type volume
                set entity_type_lower surface                
            }            
            set volume_box [Voro::GetVolumeBox]
            set volume [lindex [split [lindex [GiD_Info ListMassProperties ${entity_type}s 1:end]  end] =] 1]
            if { [string is double -strict $volume] } {
                set volume_percent [expr {int($volume/$volume_box*100)}]
            } else {
                set volume_percent ?
            }
            set boundary_surface_ids [list]
            foreach surface_id [GiD_Geometry list $entity_type_lower 1:end] {
                set surface_volume_ids [Voro::GetEntityHigherEntities $entity_type_lower $surface_id]
                if { [llength $surface_volume_ids] == 1 } {
                    lappend boundary_surface_ids $surface_id
                }
            } 
            set area [lindex [split [lindex [GiD_Info ListMassProperties ${entity_type_lower}s {*}$boundary_surface_ids]  end] =] 1]
            if { [string is double -strict $area] } {
                set specific_area [expr {$area/$volume_box}]
                set specific_area [format "%.2f" $specific_area]
            } else {
                set area ?
            }
            if { $options(dimension) == 2 } {
                GidUtils::SetWarnLine [= "Voro geometry created (volume=%s%%, perimeter/area=%s 1/length unit)" $volume_percent $specific_area]
            } else {
                GidUtils::SetWarnLine [= "Voro geometry created (volume=%s%%, area/volume=%s 1/length unit)" $volume_percent $specific_area]
            }
        } else {
            GidUtils::SetWarnLine [= "Voro geometry created"]
        }
    }
}

proc Voro::Create { } {
    variable options     
    Voro::GenerateGeometry
    if { $options(collapse) } {
        Voro::Collapse
    } else {
        if { $options(local_collapse_by_volume) } {
            #in this case try also to delete small details volume by volume
            Voro::LocalCollapseByVolume
        }
    }
    if { $options(x_periodic) || $options(y_periodic) || $options(z_periodic)} {
        if { $options(create_contacts_periodicity) } {
            Voro::CreatePeriodicityContacts
        }
    }    
    if { $options(create_matrix) } {        
        Voro::CreateExtrudedEntitiesScalingInside
    } 
}

proc Voro::CopyPoints { points_original increment } {
    set points_copy [list]
    foreach point $points_original {
        lappend points_copy [MathUtils::VectorSum $point $increment]         
    }
    return $points_copy   
}

#returns the original points and the new copies in a new file
proc Voro::CopyPointsToNeighborCellsForPeriodicity { filename } {
    variable options
    set fp [open $filename r]
    if { $fp == "" } {
        return ""
    } else {
        set points_original [list]
        while { ![eof $fp] } {
            gets $fp line
            if { $line != "" } {
                lappend points_original [lrange $line 1 3]
                if { $options(radical) } {
                    set r([lindex $line 0]) [lindex $line 4]
                }
            }
        }
        close $fp
    }
    set count [llength $points_original]
    set filename_copies $filename.copies
    file delete $filename_copies
    file copy -force -- $filename $filename_copies
    set fp [open $filename_copies a]
    set increments [Voro::GetFindingIncrements]    
    foreach increment $increments {
        set i 0
        foreach point $points_original {
            incr i
            incr count
            set point_copy [MathUtils::VectorSum $point $increment]
            if { $options(radical) } {
                puts $fp "$count $point_copy $r($i)"
            } else {
                puts $fp "$count $point_copy"
            }
        }
    }
    close $fp
    return $filename_copies
}

proc Voro::GenerateGeometry { } {
    variable options     
    if { $options(random) } {
        if { $options(save_random_centers) } {
            set filename_voro $options(particles_filename)
            if { $filename_voro == "" } {
                WarnWinText "Set the filename where the random particles will be saved"
                return 1
            }
        } else {
            set filename_voro [GidUtils::GetTmpFilename .voro] 
        }
        Voro::GenerateRandomParticles $filename_voro
    } else {
        set filename_voro $options(particles_filename)
    }
    if { ![file exists $filename_voro] } {
        WarnWinText "Voro::GenerateGeometry. Input particles file '$filename_voro' not found"
        return 1
    }
        
    if { $options(x_periodic) && $options(y_periodic) && $options(z_periodic) } {
        set periodic -p
    } else {
        set periodic ""        
        foreach axis {x y z} {
            if { $options(${axis}_periodic) } {
                if { $axis == "z" && $options(dimension) != 3 } {
                    #dimension 2.5 mean a 3D shape but prismatic (extrussion of a 2D shape)
                    continue
                }
                lappend periodic -p${axis}                        
            }   
        }
    }
    if { $options(radical) } {
        set radical -r
    } else {
        set radical ""
    }
    set limits [list $options(x_min) $options(x_max) $options(y_min) $options(y_max) $options(z_min) $options(z_max)]    
    if { $options(dimension) != 3 } {
        #dimension 2.5 mean a 3D shape but prismatic (extrussion of a 2D shape)
        lset limits 4 0
        lset limits 5 1
    }
    set exe [Voro::GetExe]  
    if { ![file exists $exe] } {
        WarnWinText "Voro::GenerateGeometry. Executable file '$exe' not found"
    } elseif { ![file executable $exe] } {
        if { $::tcl_platform(platform) == "windows"} {
            WarnWinText "Voro::GenerateGeometry. file '$exe' exists but could not be executed"
        } else {
            #try to set execution permission
            file attributes $exe -permissions +x
            if { ![file executable $exe] } {
                WarnWinText "Voro::GenerateGeometry. file '$exe' exists but could not be executed"
            }
        }
    }

    
    #trick to generate a periodic box: copy the same points to all neighbor box and cut by 6 planes to have the center box
    if { $options(periodic_box) } {
        set walls "-wb $limits" ;#cut walls
        #point limits must be increased a box on each side
        set limits ""
        foreach axis {x y z} {
            if { !$options(${axis}_periodic) } {
                lappend limits $options(${axis}_min) $options(${axis}_max)
            } else {
                lappend limits [expr {2.0*$options(${axis}_min)-$options(${axis}_max)}] [expr {2.0*$options(${axis}_max)-$options(${axis}_min)}]
            }            
        }     
        if { $options(dimension) != 3 } {
            #dimension 2.5 mean a 3D shape but prismatic (extrussion of a 2D shape)
            lset limits 4 0
            lset limits 5 1
        }
        #replace the file
        set filename_voro_periodic [Voro::CopyPointsToNeighborCellsForPeriodicity $filename_voro]
        if { $options(random) && !$options(save_random_centers) } {
            file delete $filename_voro
        }
        set filename_voro $filename_voro_periodic
    } else {
        set walls ""
    }
    
    set filename_vol ${filename_voro}.vol
    file delete $filename_vol
    if { $options(create_contacts_inner) } {
        catch {exec $exe -c {%i %w %s %P %t %n} {*}$periodic {*}$radical {*}$walls {*}$limits $filename_voro} txt
    } else {
        catch {exec $exe -c {%i %w %s %P %t} {*}$periodic {*}$radical {*}$walls {*}$limits $filename_voro} txt        
    }
    if { $options(random) } {
        #delete the axiliary file that was created random
        file delete $filename_voro    
    }
   
    Voro::ReadVoronoi $filename_vol
    file delete $filename_vol
    return 0
}

#round near zero values to zero
proc Voro::RoundNearZeroToZero { xyz } {
    set xyz_out [list]
    foreach item [split $xyz ,] {
        if { [expr abs($item)] < 1e-10 } {
            set item 0
        }
        lappend xyz_out $item
    }
    return [join $xyz_out ,]
}


proc Voro::CreateLayerNoComplain { layer } {
    if { [lsearch [GiD_Info Layers] $layer] == -1 } {
        GiD_Process 'Layers New $layer escape
    }
}

proc Voro::LayerHasEntities { layer } {
    set has_entities 0
    foreach type { points lines surfaces volumes nodes elements } {
        if { [GiD_Info Layers -count -entities $type $layer] } {
            set has_entities 1
            break
        }
    }
    return $has_entities
}

proc Voro::DeleteLayerNoComplain { layer } {
    if { [lsearch [GiD_Info Layers] $layer] != -1 } {        
        if { ![Voro::LayerHasEntities $layer] } {
            GiD_Process 'Layers Delete $layer escape
        }
    }
}

proc Voro::IsContactVolume { volume_id } {
    set is_contact_volume 0
    if { [lindex [GiD_Info list_entities volume $volume_id] 0] == "CONTACTVOLUME" } {
        set is_contact_volume 1
    }
    return $is_contact_volume
}

proc Voro::ReadFile { filename  } {
    set content ""
    if { [file exists $filename] } {
        set fp [open $filename r]
        if { $fp != "" } {
            set content [read -nonewline $fp]
            close $fp
        }
    }
    return $content
}

#filename is the .vol filename
proc Voro::ReadVoronoi { filename } {
    variable options         
    set fail 0
    if { ![file exists $filename] } {
        set fail 1
        WarnWinText "Voro::ReadVoronoi: file '$filename' not found"
    } else {
        set data [Voro::ReadFile $filename]    
        if { $data == "" } {
            set fail 1
        } else {
            if { $options(dimension) == 2 } {
                set entity_type surface
            } else {
                set entity_type volume                
            }
            Voro::CreateLayerNoComplain ${entity_type}s
            set layer_num [GiD_Info geometry maxnum${entity_type}s]
            foreach line [split $data \n] {
                set line [string trim $line]
                if { $line == "" } {
                    continue
                }
                if { !$options(share) } {
                    #unset to create them again for each polygon
                    array unset points                    
                }
                lassign [lrange $line 0 2] polygon_id num_polygon_vertex num_polygon_faces
                incr layer_num
                set layer ${entity_type}s//$layer_num                
                Voro::CreateLayerNoComplain $layer                
                set pos_first_vertex 3
                set pos_last_vertex [expr {$num_polygon_vertex+$pos_first_vertex-1}]
                set coordinates_list [lrange $line $pos_first_vertex $pos_last_vertex]
                foreach item $coordinates_list {
                    set xyz [string range $item 1 end-1]
                    set xyz [Voro::RoundNearZeroToZero $xyz]
                    set key $xyz
                    if { ![info exists points($key)] } {
                        if { $options(dimension) == 2 && [lindex [split $xyz ,] 2] > 0.5 } {
                            set points($key) 0
                        } else {
                            set points($key) [GiD_Geometry create point append $layer {*}[split $xyz ,]]
                        }
                    }
                }
                
                set pos_first_face [expr $pos_last_vertex+1]
                set pos_last_face [expr {$num_polygon_faces+$pos_first_face-1}]
                set faces_list [lrange $line $pos_first_face $pos_last_face]   
               
                set surfaces_and_versos [list]     
                array unset projected_points_face             
                set i_face 0         
                foreach item $faces_list {
                    set point_ids [list]
                    foreach local_node_id [split [string range $item 1 end-1] ,] {
                        set xyz [string range [lindex $coordinates_list $local_node_id] 1 end-1]
                        set xyz [Voro::RoundNearZeroToZero $xyz]
                        set key $xyz
                        lappend point_ids $points($key)
                    }
                    if { $options(dimension) == 2 && [lsearch $point_ids 0] != -1 } {
                        #do not create this surface in 2D
                        if { $options(create_contacts_inner) } {
                            #but find its projected line to find contacts based on neighboring information
                            set projected_point_ids [list]
                            foreach point_id $point_ids {
                                if { $point_id != 0 } {
                                    lappend projected_point_ids $point_id
                                }
                            }
                            set projected_points_face($i_face) $projected_point_ids
                        }
                        incr i_face                        
                        continue
                    }
                    set num_face_edges [llength $point_ids]
                    lappend point_ids [lindex $point_ids 0]                   
                    set lines_and_versos [list]
                    for {set i_edge 0} {$i_edge<$num_face_edges} {incr i_edge} {
                        set p0 [lindex $point_ids $i_edge]
                        set p1 [lindex $point_ids $i_edge+1]
                        set key "$p0,$p1"
                        set key_inverse "$p1,$p0"
                        if { [info exists lines($key)] } {
                            lappend lines_and_versos [list $lines($key) 0]
                        } elseif {[info exists lines($key_inverse)] } {
                            lappend lines_and_versos [list $lines($key_inverse) 1]
                        } else {
                            set lines($key) [GiD_Geometry create line append stline $layer $p0 $p1]                            
                            lappend lines_and_versos [list $lines($key) 0]
                        }
                    }
                    if { $options(dimension) == 2 } {
                        set key $polygon_id
                    } else {
                        set key [list]
                        foreach item $lines_and_versos {
                            lappend key [lindex $item 0]
                        }
                        set key [join [lsort -integer -increasing $key] ,]
                    }                    
                    if { ![info exists surfaces($key)] } {                        
                        if { $options(try_fix_gid_bug_first_verso_1) } {                           
                            if { [lindex [lindex $lines_and_versos 0] 1] == 1 } {
                                #try to avoid create a planar surface with first curve orientation==1
                                set pos [lsearch -index 1 $lines_and_versos 0]
                                if { $pos != -1 } {
                                    set lines_and_versos [concat [lrange $lines_and_versos $pos end]  [lrange $lines_and_versos 0 $pos-1]]
                                }
                            }
                        }
                        set surfaces($key) [GiD_Geometry create surface append plsurface $layer $num_face_edges {*}$lines_and_versos]  
                        set verso 0                      
                    } else {
                        set verso 1
                    }
                    if { $options(dimension) == 2 } {
                        
                    } else {
                        lappend surfaces_and_versos [list $surfaces($key) $verso]
                    }
                    incr i_face
                }
                                
                if { $options(dimension) != 2 } {
                    set key $polygon_id
                    if { ![info exists volumes($key)] } {
                        set volumes($key) [GiD_Geometry create volume append $layer $num_polygon_faces {*}$surfaces_and_versos]
                    }
                }
                                               
                if { $options(create_contacts_inner) } {                    
                    set pos_first_neighbor [expr $pos_last_face+1]
                    set pos_last_neighbor [expr {$num_polygon_faces+$pos_first_neighbor-1}]
                    set neighbor_list [lrange $line $pos_first_neighbor $pos_last_neighbor]   
                    set i_face 0
                    foreach neighbor_polygon_id $neighbor_list {
                        if { $neighbor_polygon_id > 0 } {
                            if { $options(dimension) == 2 } {
                                set projected_point_ids $projected_points_face($i_face)                                        
                                if { [llength $projected_point_ids] == 2 } {
                                    #in 2D this store line that represent the face projected int the i_face index
                                    lassign $projected_point_ids p0 p1
                                    if { [info exists lines($p0,$p1)] } {
                                        set entity_and_verso [list $lines($p0,$p1) 0]
                                    } elseif {[info exists lines($p1,$p0)] } {
                                        set entity_and_verso [list $lines($p1,$p0) 1]
                                    } else {
                                        set entity_and_verso [list -1 -1]
                                        W "Voro::ReadVoronoi. line with points $key not found"
                                    }
                                } elseif { [llength $projected_point_ids] == 0 } {
                                    #it is the 3D top surface
                                    W "Voro::ReadVoronoi. it is the 3D top surface"
                                } else {
                                    W "Voro::ReadVoronoi. expected two points: $projected_point_ids"
                                }                                  
                            } else {
                                set entity_and_verso [lindex $surfaces_and_versos $i_face]
                            }
                            #same key_neighbor A-B and B-A
                            if { $polygon_id < $neighbor_polygon_id } {
                                set key_neighbor $polygon_id,$neighbor_polygon_id
                            } else {
                                set key_neighbor $neighbor_polygon_id,$polygon_id
                            }
                            lappend neighbor_surfaces_and_versos($key_neighbor) $entity_and_verso
                        } else {
                            #periodicity neighbors (-periodic neighbor id) or computational region neighbors (-1 to -6)
                        }
                        incr i_face
                    }
                }                
            }
            if { $options(create_contacts_inner) } {
                set layer Contacts_inner
                Voro::CreateLayerNoComplain $layer 
                foreach key [array names neighbor_surfaces_and_versos] {
                    if { [llength $neighbor_surfaces_and_versos($key)] == 2 } {
                        set surfaces_and_versos $neighbor_surfaces_and_versos($key)
                        lset surfaces_and_versos {0 1} 0
                        lset surfaces_and_versos {1 1} 1
                        if { $options(dimension) == 2 } {
                            #GiD_Geometry syntax is diferent for contactsurface and contactvolume !!
                            GiD_Geometry create $entity_type append contactsurface $layer 2 {*}$surfaces_and_versos
                        } else {
                            GiD_Geometry create $entity_type append $layer 2 {*}$surfaces_and_versos contactvolume {}
                        }
                    } else {
                        WarnWinText "Expected two neighbor surfaces: $neighbor_surfaces_and_versos($key)"
                    }
                }
            }
            if { $options(convert_surfaces_to_nurbs) } {
                set surface_ids [list]
                foreach {key value} [array get surfaces] {
                    lappend surface_ids $value
                }
                set surface_ids [lsort -integer -unique $surface_ids]
                GiD_Process Mescape Geometry Edit ConvToNurbsS {*}$surface_ids escape yes escape escape
            }
            Voro::DeleteLayerNoComplain Layer0            
        }
    }
    return $fail
}

proc Voro::Collapse { } {
    variable options
    set prev_AutoImportTolerance [GiD_Set AutoImportTolerance]
    set prev_ImportTolerance [GiD_Set ImportTolerance]
    GiD_Set AutoImportTolerance 0
    GiD_Set ImportTolerance $options(tolerance)
    GiD_Process MEscape Utilities Collapse Model Yes escape
    GiD_Set AutoImportTolerance $prev_AutoImportTolerance
    GiD_Set ImportTolerance $prev_ImportTolerance
}

proc Voro::LocalCollapseByVolume { } {
    variable options
    set prev_AutoImportTolerance [GiD_Set AutoImportTolerance]
    set prev_ImportTolerance [GiD_Set ImportTolerance]
    GiD_Set AutoImportTolerance 0
    GiD_Set ImportTolerance $options(tolerance)
    foreach volume_id [GiD_Geometry list volume 1:end] {
        if { ![Voro::IsContactVolume $volume_id] } {
            GiD_Process MEscape Utilities Collapse Volumes $volume_id escape
        }
    }    
    GiD_Set AutoImportTolerance $prev_AutoImportTolerance
    GiD_Set ImportTolerance $prev_ImportTolerance
}

proc Voro::GetVolumeBox { } {
    variable options
    if { $options(dimension) == 2 } {
        set axes {x y}
    } else {
        set axes {x y z}
    }
    set volume_box 1.0
    foreach axis $axes {        
        set volume_box [expr {$volume_box*($options(${axis}_max)-$options(${axis}_min))}] 
    }
    return $volume_box
}

#type must be Points, Lines, Surfaces
proc Voro::GetBoundaryEntities { entity_type } {
    GidUtils::DisableGraphics
    GiD_Process 'SelectEntities ${entity_type}s
    GiD_Process Filter:HigherEntity=1
    GiD_Process 1:end Escape
    GidUtils::EnableGraphics
    return [GidUtils::UnCompactNumberList [lrange [string range $::GidPriv(selection) 1 end-1] 1 end]]
}

proc Voro::GetEntityCenter { entity_type entity_id } {
    if { $entity_type == "point" } {
        set center [lrange [GiD_Geometry get point $entity_id] 1 3]
    } elseif { $entity_type == "line" } {
        set center [GiD_Info parametric line $entity_id coord 0.5]
    } else {
        #planar surfaces are not parametric
        #set center [GiD_Info parametric surface $surface_id coord 0.5 0.5]
        set data [GiD_Info list_entities ${entity_type}s $entity_id]
        if { ![regexp {Center: ([^\n]*)\n} $data dummy center] } {
            WarnWinText "Error Voro::GetEntityCenter $entity_type $entity_id"
            set center {-1 -1 -1}
        }    
    } 
    return $center
}


proc Voro::GetEntityHigherEntities { entity_type entity_id } {
    set data [GiD_Info list_entities -more ${entity_type}s $entity_id]
    if { ![regexp {Higher entities (.*): ([^\n]*)\n} $data dummy entity_type_parent higher_entities] } {
        WarnWinText "Error Voro::GetEntityHigherEntities $entity_type $entity_id"
        set higher_entities ""
    }        
    return [string trim $higher_entities]
}

proc Voro::GetEntityLowerEntities { entity_type entity_id } {    
    set data [GiD_Geometry get $entity_type $entity_id]
    if { $entity_type == "volume" } {
        set lower_entities_ids [list]
        foreach item [lrange $data 2 end] {
            lappend lower_entities_ids [lindex $item 0]
        }
    } elseif { $entity_type == "surface" } {
        set num_lines [lindex $data 2]     
        set lower_entities_ids [list]  
        if { [lindex $data 0] == "nurbssurface" } {
            set first_line_index 9
        } elseif { [lindex $data 0] == "planarsurface" } {
            set first_line_index 3
        } else {
            set first_line_index 3
        }
        foreach item [lrange $data $first_line_index [expr {$first_line_index+$num_lines-1}]] {
            lappend lower_entities_ids [lindex $item 0]
        }
    } elseif { $entity_type == "line" } {
        set lower_entities_ids [lrange $data 2 3]
    } else {
        W "Voro::GetEntityLowerEntities. unexpected $entity_type"
    }
    return $lower_entities_ids
}

#to find periodic slave-surface
proc Voro::GetFindingIncrements { } {
    variable options
    #possible directions all cell neigbors 
    #6 face-neighbor, 12 edge-neighbor, 8 vertex-neigbor =26 neigbors    
    foreach axis {x y z} {
        if { !$options(${axis}_periodic) } {
            set index_min($axis) 0
            set index_max($axis) 0
        } else {
            set index_min($axis) -1
            set index_max($axis) 1
        }
        set box_size($axis) [expr {$options(${axis}_max)-$options(${axis}_min)}] 
    }
    if { $options(dimension) != 3 } {
        #dimension 2.5 mean a 3D shape but prismatic (extrussion of a 2D shape)
        set index_min(z) 0
        set index_max(z) 0
        set box_size(z) 0
    }
    set increments [list]
    for {set i $index_max(x)} {$i>=$index_min(x)} {incr i -1} {
        set x [expr {$i*$box_size(x)}]
        for {set j $index_max(y)} {$j>=$index_min(y)} {incr j -1} {
            set y [expr {$j*$box_size(y)}]
            for {set k $index_max(z)} {$k>=$index_min(z)} {incr k -1} {
                set z [expr {$k*$box_size(z)}]
                if { $i  || $j || $k} {                    
                    lappend increments [list $x $y $z]
                }
            }
        }
    }
    return $increments
}

#3D case: create separated contact volumes
#2D case: create separated contact surfaces
proc Voro::CreatePeriodicityContacts { } {
    variable options
    variable advancebar
    set tolerance2 [expr {$options(tolerance)*$options(tolerance)}]
    package require vtk    
    if { $options(dimension) == 2 } {
        set entity_type line
        set contact_level_type surface
        set contact_type contactsurface
    } else {
        set entity_type surface
        set contact_level_type volume
        set contact_type contactvolume
    }
    set boundary_entities [Voro::GetBoundaryEntities $entity_type]
    
    set vtkPoints [vtkPoints New]
    foreach entity_id $boundary_entities {
        set entity_center($entity_id) [format {%.4f %.4f %.4f} {*}[Voro::GetEntityCenter $entity_type $entity_id]]
        #truncate a few to have more coincidences: center(A) == center(B)+periodic_displacement
        set vtk_id [$vtkPoints InsertNextPoint {*}$entity_center($entity_id)]
        set entity_id_kdtree($vtk_id) $entity_id
    }     
    #create a kdtree for fast spatial search   
    set vtkKdTree [vtkKdTree New]
    $vtkKdTree SetTolerance $options(tolerance)
    $vtkKdTree BuildLocatorFromPoints $vtkPoints
    
    #get allowed neighbor directions (3D: 26 if periodic in x y z, 2D: 8 if periodic in x y)
    set Voro::advancebar(percent) 0.0
    set Voro::advancebar(stop) 0
    GidUtils::CreateAdvanceBar [= "Finding periodicities"] [= "Percentage"] Voro::advancebar(percent) 100  {set Voro::advancebar(stop) 1}
    set increments [Voro::GetFindingIncrements]     
    set num_boundary_entities [llength $boundary_entities]
    set counter 0
    set num_found_brute_force 0
    foreach entity_id $boundary_entities { 
        incr counter
        if { [info exists used_entity($entity_id)]} {
            continue
        }     
        if { ![expr {$counter%5}] } {
            set Voro::advancebar(percent) [expr {int(double([array size used_entity])/$num_boundary_entities*100.0)}]
            if { $Voro::advancebar(stop) } {
                WarnWin [= "User stop"]
                break
            }
        }
        foreach increment $increments {             
            set translated_center [MathUtils::VectorSum $entity_center($entity_id) $increment]                    
            #set vtk_id [$vtkKdTree FindClosestPoint {*}$translated_center d2]
            #FindClosestPoint not exists in the tcl wrapped command !!
            set vtk_id [$vtkKdTree FindPoint {*}$translated_center]
            if { $vtk_id != -1 } {
                set id $entity_id_kdtree($vtk_id)
                set slave_entity($entity_id) $id
                set translation_entity($entity_id) $increment
                set used_entity($entity_id) 1
                set used_entity($id) 1
                break
            }            
        }
        
        if { ![info exists used_entity($entity_id)] } {
            foreach increment $increments {             
                set translated_center [MathUtils::VectorSum $entity_center($entity_id) $increment]
                #O(n^2) search, too expensive for big n, could use octree, bin, etc to reach O(n log(n))
                foreach id $boundary_entities {
                    if { $id != $entity_id && ![info exists used_entity($id)]} {
                        set d2 [MathUtils::VectorDistance2 $translated_center $entity_center($id)]
                        if { $d2 < $tolerance2 } {
                            incr num_found_brute_force
                            set slave_entity($entity_id) $id
                            set translation_entity($entity_id) $increment
                            set used_entity($entity_id) 1
                            set used_entity($id) 1
                            break
                        }
                    }
                }
                if { [info exists used_entity($entity_id)]} {
                    #increment found
                    break
                }
            }
        }
    }
    GidUtils::DeleteAdvanceBar
    $vtkKdTree Delete
    $vtkPoints Delete
#     if { $num_found_brute_force } {
#         WarnWinText "found by brute force $num_found_brute_force of [array size used_entity]"
#     }
        
    set unmatched_entities [list]
    foreach entity_id $boundary_entities {
        if { ![info exists used_entity($entity_id)] } {           
            lappend unmatched_entities $entity_id                      
        }
    }
    if { [llength $unmatched_entities] } {
        if { $options(x_periodic) && $options(y_periodic) && $options(z_periodic)} {
            #raise only error if periodic in all directions, else is normal to have unmatched entities
            WarnWinText "Error: [llength $unmatched_entities] unmatched entities"
            WarnWinText "unmatched: $unmatched_entities"
        }          
    }
    set layer Contacts_outer
    Voro::CreateLayerNoComplain $layer
    foreach entity_id [lsort -integer [array names slave_entity]] {
        #GiD_Process Mescape Geometry Create Contact SeparatedVolume $entity_id $slave_entity($entity_id) escape
        set entities_and_versos [list [list $entity_id 0] [list $slave_entity($entity_id) 1]]
        lassign $translation_entity($entity_id) translation_x translation_y translation_z
        set transformation_matrix [list 1 0 0 $translation_x 0 1 0 $translation_y 0 0 1 $translation_z 0 0 0 1]        
        if { $options(dimension) == 2 } {
            W "GiD_Geometry create $contact_level_type append $layer 2 {*}$entities_and_versos $contact_type $transformation_matrix"
        } else {  
            GiD_Geometry create $contact_level_type append $layer 2 {*}$entities_and_versos $contact_type $transformation_matrix
        }
    }        

    if { $options(create_contacts_periodicity_element_type) == "None" } {
        #could set contact volumes to not generate any element, only to force equivalent mesh in both entities to have periodicity
        GiD_Process MEscape Meshing MeshCriteria NoMesh ${contact_level_type}s Layer:$layer escape
    } elseif { $options(create_contacts_periodicity_element_type) == "Line" } {        
        #set to mesh these contact volumes with line elements linking periodic nodes        
        GiD_Process MEscape Meshing ElemType Linear ${contact_level_type}s Layer:$layer escape
    } elseif { $options(create_contacts_periodicity_element_type) == "Prism" } {
        #default behavior, create prisms (or hexahedra if entities were meshed with cuadrilaterals)
        #                         quadrilateral for contact surfaces
        #GiD_Process MEscape Meshing ElemType Prism ${contact_level_type}s Layer:$layer escape
    } else {
        WarnWinText "Error unexpected case $options(create_contacts_periodicity_element_type)"
    }    
}

proc Voro::OptionsWindow { } {    
    variable info_plugin
    variable options
    if { ![info exists options(num_random_centers)] } {
        Voro::OptionsReset
    }

    set w .gid.voro
    InitWindow $w "Voro++ $info_plugin(Version)" VoroWindowGeom   
    set f [ttk::frame $w.f]
    

    ttk::labelframe $f.fdimension -text [= "Dimension"]
    ttk::radiobutton $f.fdimension.2d -text 2D -value 2 -variable Voro::options(dimension)
    ttk::radiobutton $f.fdimension.2_5d -text 2.5D -value 2.5 -variable Voro::options(dimension)
    ttk::radiobutton $f.fdimension.3d -text 3D -value 3 -variable Voro::options(dimension)
    GidHelp $f.fdimension [= "Select 2D a 3D prismatic or a full 3D case"]   

    ttk::labelframe $f.fparticles -text [= "Polygon centers"]
    ttk::radiobutton $f.fparticles.random -text [= "Random"] -value 1 -variable Voro::options(random)
    ttk::radiobutton $f.fparticles.fromfile -text [= "From file"] -value 0 -variable Voro::options(random)
    GidHelp $f.fparticles.random [= "To generate polygon centers randomly"]   
    ttk::label $f.fparticles.l_num_random_centers -text [= "Number or centers"]
    ttk::entry $f.fparticles.e_num_random_centers -textvariable Voro::options(num_random_centers)
    ttk::checkbutton $f.fparticles.c_save_random_centers -text [= "Save centers to file"] -variable Voro::options(save_random_centers)
    GidHelp $f.fparticles.e_num_random_centers [= "Set the number or random polygon centers to be generated"]      
    ttk::label $f.fparticles.l_filename -text [= "Particles file name"]
    ttk::entry $f.fparticles.e_filename -textvariable Voro::options(particles_filename)
    ttk::button $f.fparticles.b_filename -image [gid_themes::GetImage "folder.png"] -command [list Voro::SelectParticlesFilename]
    GidHelp [list $f.fparticles.b_filename $f.fparticles.e_filename] [= "To read polygon centers from a file (a text file with id x y z ?radius?)"]      
    ttk::checkbutton $f.fparticles.radical -text [= "Radical"] -variable Voro::options(radical)   
    GidHelp $f.fparticles.radical [= "The Voronoi tessellation take into account the particle radius as weigth to calculate its polygon (ramdom radius were generated also)"]      
    
    ttk::labelframe $f.flimits -text [= "Limits"]
    foreach axis {x y z} {
        foreach side {min max} txt [list [= "min"] [= "max"]] {
            ttk::label $f.flimits.l_limit_${axis}_${side} -text "$axis $txt"
            ttk::entry $f.flimits.e_limit_${axis}_${side} -textvariable Voro::options(${axis}_${side})
            GidHelp $f.flimits.e_limit_${axis}_${side} [= "Set the box limits"]           
        }
    }

    ttk::labelframe $f.fperiodic -text [= "Periodicity"]
    ttk::frame $f.fperiodic.faxis
    foreach axis {x y z} {        
        ttk::checkbutton $f.fperiodic.faxis.periodic_${axis} -text [concat ${axis} [= "periodic"]] -variable Voro::options(${axis}_periodic)
        GidHelp $f.fperiodic.faxis.periodic_${axis} [= "To force the periodicity of the shape in the boundary"]
    }
    ttk::frame $f.fperiodic.fmore
    ttk::checkbutton $f.fperiodic.fmore.periodic_box -text [= "Force box shape"] -variable Voro::options(periodic_box)
    GidHelp $f.fperiodic.fmore.periodic_box [= "Force the shape to be a box. The amount of volumes will be higher than the provided centers"]

    ttk::checkbutton $f.fperiodic.fmore.contacts -text [= "Force mesh periodicity"] -variable Voro::options(create_contacts_periodicity)
    GidHelp $f.fperiodic.fmore.contacts [= "Create 'separated contact volumes' between outer periodic surfaces, to force mesh periodicity"]
    
    ttk::label $f.fperiodic.fmore.lcontacts_element_type -text [= "Link element"]
    ttk::combobox $f.fperiodic.fmore.contacts_element_type -values [list "None" "Line" "Prism"] -textvariable Voro::options(create_contacts_periodicity_element_type)
    GidHelp $f.fperiodic.fmore.contacts_element_type [= "Element type in periodicity contacts"]
   
    ttk::labelframe $f.fgrain_boundary -text [= "Grain boundary"]
    ttk::checkbutton $f.fgrain_boundary.contacts_inner -text [= "Create contacts"] -variable Voro::options(create_contacts_inner)       
    GidHelp $f.fgrain_boundary.contacts_inner [= "Create contact volumes between each two inner surfaces"]      
    ttk::checkbutton $f.fgrain_boundary.create_matrix -text [= "Create matrix"] -variable Voro::options(create_matrix)       
    GidHelp $f.fgrain_boundary.create_matrix [= "Create the volumes of the matrix contracting the grains"]    
    ttk::checkbutton $f.fgrain_boundary.create_matrix_delete_outer -text [= "Delete matrix on boundary"] -variable Voro::options(create_matrix_delete_outer)
    GidHelp $f.fgrain_boundary.create_matrix_delete_outer [= "To delete or not the matrix part touching the outer box"]
    ttk::label $f.fgrain_boundary.l_contraction_factor -text [= "Contraction factor"]
    ttk::entry $f.fgrain_boundary.e_contraction_factor -textvariable Voro::options(create_matrix_scale_factor)
    GidHelp $f.fgrain_boundary.l_contraction_factor [= "factor from 0.0 to 1.0. (approx volume to get: 0.85->24%%, 0.91->15%%, 0.94 ->10%%)"]
    ttk::checkbutton $f.fgrain_boundary.create_matrix_create_grain -text [= "Create grain"] -variable Voro::options(create_matrix_create_grain)
    GidHelp $f.fgrain_boundary.create_matrix_create_grain [= "To create or not the contracted grain volumes"]
    
    
    ttk::labelframe $f.fcollapse -text [= "Join"]
    ttk::checkbutton $f.fcollapse.share -text [= "Share"] -variable Voro::options(share)
    GidHelp $f.fcollapse.share [= "Entities are created shared between neighbors"]
    ttk::checkbutton $f.fcollapse.collapse -text [= "Collapse"] -variable Voro::options(collapse)
    GidHelp $f.fcollapse.collapse [= "Join entities closer than the tolerance (to remove very small details)"]
    ttk::label $f.fcollapse.l_tolerance -text [= "Tolerance"]
    ttk::entry $f.fcollapse.e_tolerance -textvariable Voro::options(tolerance)
    GidHelp $f.fcollapse.e_tolerance  [= "Set the tolerance to join nearby points and to find surface pairs to create contacts"]           
    
    #lower buttons
    ttk::frame $w.frmButtons -style BottomFrame.TFrame    
    ttk::button $w.frmButtons.btngenerate -text [= "Create"] -command Voro::CreateAsk -underline 0 -style BottomFrame.TButton
    ttk::button $w.frmButtons.btnclose -text [= "Close"] -command [list destroy $w] -underline 0 -style BottomFrame.TButton
   
    grid $f.fdimension.2d $f.fdimension.2_5d $f.fdimension.3d -sticky w
    grid $f.fdimension -sticky ew -columnspan 2
    grid columnconfigure $f.fdimension 3 -weight 1
    
    grid $f.fparticles.random $f.fparticles.fromfile -sticky ew
    grid $f.fparticles.l_num_random_centers $f.fparticles.e_num_random_centers -sticky w    
    grid configure $f.fparticles.e_num_random_centers -sticky ew -columnspan 2
    grid $f.fparticles.c_save_random_centers -sticky w -columnspan 3
    grid $f.fparticles.l_filename $f.fparticles.e_filename $f.fparticles.b_filename -sticky w
    grid configure $f.fparticles.e_filename -sticky ew
    grid $f.fparticles.radical -sticky w -columnspan 3
    grid $f.fparticles -sticky ew -columnspan 2
    grid columnconfigure $f.fparticles 1 -weight 1

    foreach axis {x y z} {
        foreach side {min max} {
            grid $f.flimits.l_limit_${axis}_min $f.flimits.e_limit_${axis}_min $f.flimits.l_limit_${axis}_max $f.flimits.e_limit_${axis}_max -sticky w
        }
    }
    grid $f.flimits -sticky ew -columnspan 2
    grid columnconfigure $f.flimits {1 3} -weight 1

    foreach axis {x y z} {
        grid $f.fperiodic.faxis.periodic_${axis} -sticky w
    }
    grid $f.fperiodic.fmore.periodic_box -sticky w -columnspan 2
    grid $f.fperiodic.fmore.contacts -sticky w -columnspan 2
    grid $f.fperiodic.fmore.lcontacts_element_type $f.fperiodic.fmore.contacts_element_type -sticky w
    grid configure $f.fperiodic.fmore.contacts_element_type -sticky ew
    grid $f.fperiodic.faxis $f.fperiodic.fmore -sticky nsew
    grid columnconfigure $f.fperiodic.faxis 0 -weight 1
    grid columnconfigure $f.fperiodic.fmore 1 -weight 1
    grid $f.fperiodic -sticky ew -columnspan 2
    grid columnconfigure $f.fperiodic {0 1} -weight 1

    grid $f.fgrain_boundary.contacts_inner -sticky w -columnspan 2
    grid $f.fgrain_boundary.create_matrix $f.fgrain_boundary.create_matrix_delete_outer  -sticky w
    grid $f.fgrain_boundary.l_contraction_factor $f.fgrain_boundary.e_contraction_factor -sticky w
    grid configure $f.fgrain_boundary.e_contraction_factor -sticky ew
    grid $f.fgrain_boundary.create_matrix_create_grain -sticky w -columnspan 2
    grid $f.fgrain_boundary -sticky ew -columnspan 2
    grid columnconfigure $f.fgrain_boundary {1} -weight 1
        
    grid $f.fcollapse.share -sticky w -columnspan 2
    grid $f.fcollapse.collapse -sticky w -columnspan 2
    grid $f.fcollapse.l_tolerance $f.fcollapse.e_tolerance -sticky w
    grid $f.fcollapse -sticky ew -columnspan 2
    grid columnconfigure $f.fcollapse 1 -weight 1

    grid $f -sticky nsew -padx 3 -pady 3
    grid columnconfigure $f 1 -weight 1
    grid rowconfigure $f 10 -weight 1

    grid $w.frmButtons -sticky ews -columnspan 7
    grid $w.frmButtons.btngenerate $w.frmButtons.btnclose -padx 3 -pady 3
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1
    if { $::tcl_version >= 8.5 } { grid anchor $w.frmButtons center }

    trace add variable Voro::options(dimension) write Voro::OnChangeDimension
    trace add variable Voro::options(random) write Voro::OnChangeRandom
    trace add variable Voro::options(save_random_centers) write Voro::OnChangeSaveRandomCenters
    foreach axis {x y z} {
        trace add variable Voro::options(${axis}_periodic) write Voro::OnChangePeriodic
    }
    trace add variable Voro::options(create_contacts_periodicity) write Voro::OnChangeCreateContactsPeriodic    
    trace add variable Voro::options(create_matrix) write Voro::OnChangeCreateMatrix    
       
    #invoke OnChange-xx to update the state depending on current values
    Voro::OnChangeDimension
    Voro::OnChangeRandom
    Voro::OnChangeSaveRandomCenters 
    Voro::OnChangePeriodic
    Voro::OnChangeCreateContactsPeriodic
    Voro::OnChangeCreateMatrix
   
    focus $w.frmButtons.btnclose
    bind $w <Alt-c> "$w.frmButtons.btnclose invoke"
    bind $w <Escape> "$w.frmButtons.btnclose invoke"
    bind $w <Destroy> [list +Voro::DestroyOptionsWindow %W $w] ;# + to add to previous script
}

proc Voro::DestroyOptionsWindow { W w } {
    variable options
    if { $W != $w } return   
    trace remove variable Voro::options(random) write Voro::OnChangeRandom
    trace remove variable Voro::options(save_random_centers) write Voro::OnChangeSaveRandomCenters
    foreach axis {x y z} {
        trace remove variable Voro::options(${axis}_periodic) write Voro::OnChangePeriodic
    }    
    trace remove variable Voro::options(create_contacts_periodicity) write Voro::OnChangeCreateContactsPeriodic
    trace remove variable Voro::options(create_matrix) write Voro::OnChangeCreateMatrix
}

proc Voro::OnChangeDimension { args } {
    variable options    
    set f .gid.voro.f
    if { [winfo exists $f] } {
        if { $options(dimension) == 2 } {
            set axis z
            foreach side {min max} {
                grid remove $f.flimits.l_limit_${axis}_min $f.flimits.e_limit_${axis}_min $f.flimits.l_limit_${axis}_max $f.flimits.e_limit_${axis}_max
            }
            grid remove $f.fperiodic.faxis.periodic_${axis}
            
            $f.fperiodic.fmore.contacts_element_type configure -values [list "None" "Line" "Quadrilateral"]
            if { $options(create_contacts_periodicity_element_type) == "Prism" } {
                set options(create_contacts_periodicity_element_type) "Quadrilateral"
            }
        } else {
            set axis z
            foreach side {min max} {
                grid $f.flimits.l_limit_${axis}_min $f.flimits.e_limit_${axis}_min $f.flimits.l_limit_${axis}_max $f.flimits.e_limit_${axis}_max
            }
            grid $f.fperiodic.faxis.periodic_${axis} -sticky w
            $f.fperiodic.fmore.contacts_element_type configure -values [list "None" "Line" "Prism"]
            if { $options(create_contacts_periodicity_element_type) == "Quadrilateral" } {
                set options(create_contacts_periodicity_element_type) "Prism"
            }
        }
    }
}

proc Voro::OnChangeRandom { args } {
    variable options    
    set f .gid.voro.f
    if { [winfo exists $f] } {
        if { $options(random) } {
            $f.fparticles.e_num_random_centers configure -state normal
            $f.fparticles.c_save_random_centers configure -state normal
            if { $options(save_random_centers) } {
                $f.fparticles.e_filename configure -state normal
                $f.fparticles.b_filename configure -state normal
            } else {
                $f.fparticles.e_filename configure -state disabled
                $f.fparticles.b_filename configure -state disabled
            }
        } else {
            $f.fparticles.e_num_random_centers configure -state disabled   
            $f.fparticles.c_save_random_centers configure -state disabled
            $f.fparticles.e_filename configure -state normal
            $f.fparticles.b_filename configure -state normal
        }
    }
}

proc Voro::OnChangeSaveRandomCenters { args } {
    variable options
    set f .gid.voro.f
    if { [winfo exists $f] } {
        if { $options(save_random_centers) } {
            $f.fparticles.e_filename configure -state normal
            $f.fparticles.b_filename configure -state normal
        } else {
            $f.fparticles.e_filename configure -state disabled
            $f.fparticles.b_filename configure -state disabled
        }
    }
}

proc Voro::SelectParticlesFilename { } {
    variable options
    set types [list [list Voro++ .voro] [list [_ "All files"] .*]]
    set defaultextension ""
    set title [_ "Select file"]
    if { $options(random) } {
        if { $options(save_random_centers) } {
            set filename [Browser-ramR file write .gid $title $options(particles_filename) $types $defaultextension 0]
        } else {
            WarnWinText "The filename is not used generating random particle if 'Save centers to file' is not set"
        }
    } else {
        set filename [Browser-ramR file read .gid $title $options(particles_filename) $types $defaultextension 0]
    }
    if { $filename != "" } {
        set options(particles_filename) $filename       
    }
}

proc Voro::OnChangePeriodic { args } {
    variable options    
    set f .gid.voro.f  
    if { [winfo exists $f] } {
        if { $options(x_periodic) || $options(y_periodic) || $options(z_periodic)} {
            $f.fperiodic.fmore.periodic_box configure -state normal            
            $f.fperiodic.fmore.contacts configure -state normal
        } else {
            $f.fperiodic.fmore.periodic_box configure -state disabled            
            $f.fperiodic.fmore.contacts configure -state disabled   
        }
    }
}

proc Voro::OnChangeCreateContactsPeriodic { args } {
    variable options    
    set f .gid.voro.f  
    if { [winfo exists $f] } {
        if { $options(create_contacts_periodicity) } {
            $f.fperiodic.fmore.contacts_element_type configure -state normal
        } else {
            $f.fperiodic.fmore.contacts_element_type configure -state disabled  
        }
    }
}

proc Voro::OnChangeCreateMatrix { args } {
    variable options
    set f .gid.voro.f  
    if { [winfo exists $f] } {
        if { $options(create_matrix) }  {
            $f.fgrain_boundary.create_matrix_delete_outer configure -state normal
            $f.fgrain_boundary.e_contraction_factor configure -state normal
            $f.fgrain_boundary.create_matrix_create_grain configure -state normal            
        } else {
            $f.fgrain_boundary.create_matrix_delete_outer configure -state disabled
            $f.fgrain_boundary.e_contraction_factor configure -state disabled            
            $f.fgrain_boundary.create_matrix_create_grain configure -state disabled
        }
    }
}

proc Voro::AddToMenu { } {
    if { [GidUtils::IsTkDisabled] } {
        return
    }   
    if { $::GidPriv(HideVolumeLevel) == 1 } {
        return
    }
    if { [GiDMenu::GetOptionIndex Geometry [list Create Voronoi...] PRE] != -1 } {
        return
    }
    #try to insert this menu after the word "Mesh->Generate mesh"
    set position [GiDMenu::GetOptionIndex Geometry [list Create Object] PRE]
    if { $position == -1 } {
        set position end
    }
    GiDMenu::InsertOption Geometry [list Create Voronoi...] $position PRE Voro::OptionsWindow "" "" insertafter _   
}

 
#invoke this menus changes
Voro::AddToMenu
GiDMenu::UpdateMenus

#register the proc to be automatically called when re-creating all menus (e.g. when doing files new)
GiD_RegisterPluginAddedMenuProc Voro::AddToMenu
